from typing import List, Tuple

from sqlalchemy import create_engine
import oracledb

from enums.DatabaseDialectEnum import DatabaseDialectEnum
from enums.DatabaseTypeEnum import DatabaseTypeEnum
from components.Databases.Enums.OracleDataTypeEnum import OracleDataTypeEnum
from components.Databases.Enums.OracleDataTypeORMEnum import OracleDataTypeORMEnum
from components.Model.InsumoModel import InsumoModel, InsumoModelOracle
from sqlalchemy import Column
import logging

logger = logging.getLogger()


class Oracle:

    def __init__(self, host: str, port: int, user: str, password: str, service: str) -> None:
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.service = service
        self.arraysize = 1000
        self.engine = None
        self.connection = None

    def get_basic_connection(self):
        try:
            if isinstance(self.connection, type(None)):
                self.connection = oracledb.connect(user=self.user, password=self.password, host=self.host,
                                                   port=self.port, service_name=self.service)
        except Exception as e:
            logger.error(f"Error obteniendo conexion básica de Oracle. {e}")
        finally:
            return self.connection

    def create_engine(self) -> None:
        try:
            dialect = DatabaseDialectEnum.ORACLE.value
            url = f"{dialect}://{self.user}:{self.password}@{self.host}:{str(self.port)}?service_name={self.service}"
            self.engine = create_engine(url, arraysize=self.arraysize)
        except Exception as e:
            logger.error(f"Error creando engine de Oracle. {e}")

    def create_model(self, tablename: str, fields: List[Tuple[str]], db_target: str, modelName: str = "TableModel"):
        model = None
        try:
            db_args = {
                '__tablename__': tablename,
                '__table_args__': {'extend_existing': True}
            }
            if db_target == DatabaseTypeEnum.ORACLE.value:
                model = type(modelName, (InsumoModelOracle,), db_args)
            else:
                model = type(modelName, (InsumoModel,), db_args)
            for field in fields:
                logger.debug(f"Attribute: {field}")
                name = field[0]
                # Default precision for Integer Oracle : 38
                if not isinstance(field[2], type(None)):
                    size = int(field[2] / 4)
                    try:
                        if not isinstance(field[3], type(None)) and field[3] > 0:
                            data_type = OracleDataTypeORMEnum[OracleDataTypeEnum(field[1].name).name].value(precision=field[3], scale=field[3])
                        elif not isinstance(field[3], type(None)):

                            data_type = OracleDataTypeORMEnum[OracleDataTypeEnum(field[1].name).name].value(precision=38, scale=field[3])
                        else:
                            data_type = OracleDataTypeORMEnum[OracleDataTypeEnum(field[1].name).name].value(size)
                    except TypeError:
                        data_type = OracleDataTypeORMEnum[OracleDataTypeEnum(field[1].name).name].value()
                elif not isinstance(field[3], type(None)) and field[3] < 0:
                    data_type = OracleDataTypeORMEnum[OracleDataTypeEnum(field[1].name).name].value(precision=38, scale=16)
                elif not isinstance(field[3], type(None)) and field[3] >= 0:
                    try:
                        data_type = OracleDataTypeORMEnum[OracleDataTypeEnum(field[1].name).name].value(precision=38)
                    except TypeError:
                        data_type = OracleDataTypeORMEnum[OracleDataTypeEnum(field[1].name).name].value()
                else:
                    data_type = OracleDataTypeORMEnum[OracleDataTypeEnum(field[1].name).name].value()
                setattr(model, name, Column(data_type))
            model = model.__table__
        except Exception as e:
            logger.error(f"Error creando modelo dinámico en Oracle con nombre {tablename}. {type(e)}.{e}")
        finally:
            return model

    def generate_sql_procedure(self, command: str, reserved_word: str = "execute") -> str:
        response = ""
        try:
            command = command.replace(reserved_word, "").replace(";", "")
            response = f"begin {command}; end;"
            logger.debug("COMANDO ORACLE:", response)
        except Exception as e:
            logger.error(f"Error generando comando sql para procedure Oracle. Comando: {command}. {e}")
        finally:
            return response

    def get_all_tablenames(self) -> List[str]:
        tablenames = []
        try:
            user = self.user.upper()
            command = f"SELECT table_name FROM all_tables WHERE OWNER='{user}'"
            with self.engine.connect() as conn:
                tablenames = conn.execute(command).all()
        except Exception as e:
            logger.error(f"Error obteniendo los nombres de tablas. {e}")
        finally:
            return tablenames

    def check_procedure(self, procedure_name, connection) -> bool:
        exists = False
        try:
            check_query = f"SELECT text FROM all_source WHERE name = '{procedure_name}'"
            result = connection.execute(check_query)
            exists = bool(result.fetchone())
        except Exception as e:
            logger.error(f"Error obteniendo existencia de procedure. {e}")
        finally:
            return exists

    def check_table(self, table_name, connection) -> bool:
        exists = False
        try:
            check_query = f"DESCRIBE {table_name}"
            result = connection.execute(check_query)
            exists = bool(result.fetchone())
        except Exception as e:
            logger.error(f"Error obteniendo existencia de tabla. {e}")
        finally:
            return exists

    def verify_table(self, table_name, connection) -> bool:
        exists = False
        try:
            check_query = f"SELECT COUNT(*) FROM {table_name}"
            result = connection.execute(check_query)
            if result > 0:
                exists = True
        except Exception as e:
            logger.error(f"Error obteniendo counts de tabla. {e}")
        finally:
            return exists