from typing import Any, Dict
import os
import json
import pandas as pd
from airflow.utils.task_group import TaskGroup
from airflow.operators.python import PythonOperator
from airflow.models import Variable
from airflow.decorators import task
from airflow.exceptions import AirflowSkipException

from enums.ProcessStatusEnum import ProcessStatusEnum
from enums.DataTypeEnum import DataTypeEnum
from enums.DatabaseTypeEnum import DatabaseTypeEnum
from components.S3Route import save_df_to_s3, load_control_to_s3
from components.Utils import select_multiple, create_temp_file, delete_temp_dir
from components.Control import get_tasks_from_control, update_new_process
from components.Xcom import delete_all_xcom_tasks, delete_task_instances
from components.DatabaseOperation.DatabaseExtraction import get_iterator, get_steps
from enums.OperationTypeEnum import OperationTypeEnum
from components.Timezone import datetime_by_tzone

import logging

logger = logging.getLogger()


def validate_generate(control_params: Dict[str, Any], timezone: str, provider: str, **kwargs) -> None:
    delete_task_instances()
    ti = kwargs["ti"]
    created_tables = ti.xcom_pull(task_ids="MASTER_TRANSFORMATION", key="TABLES_CREATED")
    lista = []
    if created_tables:
        for i in created_tables:
            lista.append(i)
    success_tasks = ti.xcom_pull(task_ids="GENERATORS", key="SUCCESS_TASKS")
    failed_tasks = ti.xcom_pull(task_ids="GENERATORS", key="FAILED_TASKS")
    conf = ti.xcom_pull(task_ids="VALIDATE_TRANSFORMATION", key="CONTROL-CONFIG")
    final_dict = {}
    status = ProcessStatusEnum.SUCCESS.value
    if not isinstance(success_tasks, type(None)) and len(success_tasks) > 0:
        for success_task in success_tasks:
            task = ti.xcom_pull(task_ids="GENERATORS", key=success_task)[0]
            final_dict.update({success_task: task})
    if not isinstance(failed_tasks, type(None)) and len(failed_tasks) > 0:
        status = ProcessStatusEnum.FAIL.value
        for failed_task in failed_tasks:
            task = ti.xcom_pull(task_ids="GENERATORS", key=failed_task)[0]
            final_dict.update({failed_task: task})
    conf = update_new_process(conf, status, final_dict, timezone, ti, lista, True)
    ti.xcom_push(key="CREATED_TABLES", value=lista)
    if status == ProcessStatusEnum.FAIL.value:
        conn = control_params["connection_id"]
        bucket = control_params["bucket"]
        prefix = control_params["prefix"]
        if not prefix.endswith("/"):
            prefix += "/"
        key = prefix + control_params["filename"]
        conf = json.dumps(conf, indent=2, default=str)
        loaded = load_control_to_s3(bytes(conf.encode()), conn, bucket, key, provider, timezone)
        if loaded:
            logger.info(f"Cargado correctamente el archivo de control en {key}")
        delete_all_xcom_tasks()
        raise AirflowSkipException(f"Ocurrieron errores en la etapa de generación")
    elif status == ProcessStatusEnum.SUCCESS.value:
        ti.xcom_push(key="CONTROL-CONFIG", value=conf)


def on_failure_generator(context) -> None:
    ti = context["ti"]
    task_name = f"{ti.task_id}_{ti.map_index}"
    selects = Variable.get('GENERATES', default_var=[], deserialize_json=True)
    table = selects[ti.map_index]
    table = select_multiple(table)["tablename"]
    exception = str(context["exception"])
    status = ProcessStatusEnum.FAIL.value
    init_process = ti.xcom_pull(task_ids="TRANSFORMATIONS", key="INIT_PROCESS_DATETIME_" + str(ti.map_index))[0]
    task_result = {"description": table, "status": status, "message": exception, "init_process": init_process,
                   "end_process": ""}
    ti.xcom_push(key=task_name, value=task_result)
    ti.xcom_push(key="FAILED_TASKS", value=task_name)


def on_success_generator(context) -> None:
    ti = context["ti"]
    task_name = f"{ti.task_id}_{ti.map_index}"
    selects = Variable.get('GENERATES', default_var=[], deserialize_json=True)
    table = selects[ti.map_index]
    table = select_multiple(table)["tablename"]
    status = ProcessStatusEnum.SUCCESS.value
    init_process = ti.xcom_pull(task_ids="GENERATORS", key="INIT_PROCESS_DATETIME_" + str(ti.map_index))[0]
    end_process = ti.xcom_pull(task_ids="GENERATORS", key="END_PROCESS_DATETIME_" + str(ti.map_index))[0]
    task_result = {"description": table, "status": status, "message": "", "init_process": init_process,
                   "end_process": end_process}
    ti.xcom_push(key=task_name, value=task_result)
    ti.xcom_push(key="SUCCESS_TASKS", value=task_name)


def generate_and_deploy(command: str, intern_conn, params: Dict[str, Any], timezone: str,
                        provider: str, chunksize=10000, **kwargs):
    if isinstance(command, type(None)):
        raise AirflowSkipException
    task = kwargs['ti']
    init_process_datetime = datetime_by_tzone(timezone).strftime('%d/%m/%Y %H:%M:%S')
    task.xcom_push(key="INIT_PROCESS_DATETIME_" + str(task.map_index), value=init_process_datetime)
    engine = intern_conn.engine
    logger.debug(f"COMANDO: {command}")
    tablename = select_multiple(command)["tablename"]
    logger.info(f"Generando resultado de la tabla {tablename}")

    definitions = task.xcom_pull(task_ids="SCRIPTS-EXTRACTOR", key="EXTRACTION-DEFINITION-JSON")
    campos = {}
    for procedure in definitions:
        if procedure["identifier"] == tablename:
            for field in procedure["fields"]:
                name = field["name"]
                campos[name] = field["datatype"]

    # Creamos el archivo temporal
    filename_mask = params["filename_mask"]
    file_type = params["file_type"]
    pattern = params["datetime_pattern"]
    delimiter = params["delimiter"]
    tmp_path = params["tmp_path"]
    tmp_file = create_temp_file(tmp_path, filename_mask, file_type, tablename, timezone, pattern)
    logger.info(f"Ruta creada: {tmp_file}")
    logger.debug(f"TABLA: {tablename}")
    steps = get_steps(tablename, chunksize, engine, True)
    steps = steps["steps"]
    if intern_conn.db_type == DatabaseTypeEnum.ORACLE.value:
        tablename = f"SELECT * FROM {tablename}"
    iterator = get_iterator(tablename, chunksize, engine)
    iterator = iterator["iterator"]
    logger.info(f"Total de pasos para generar archivo resultado: {steps}")
    for step in range(steps):
        logger.debug(f"STEP: {step}")
        header = True if step == 0 else False
        try:
            dataframe = next(iterator)
            for campo in dataframe.columns:
                if campo in campos.keys():
                    if campos[campo] == DataTypeEnum.DATE.name:
                        dataframe[campo] = pd.to_datetime(dataframe[campo]).dt.date
                    # elif campos[campo] == DataTypeEnum.DATETIME.name:  # datetime:
                    #     dataframe[campo] = pd.to_datetime(dataframe[campo], format='%Y-%m-%d %H:%M:%S')

            dataframe = dataframe.drop("INTERN_ID_BCOM", axis=1, errors='ignore')
            logger.debug(dataframe)
            dataframe.to_csv(tmp_file, sep=delimiter, index=False, mode='a', header=header)
        except StopIteration:
            break
    list_outputs = params["s3_params"]
    size = os.path.getsize(tmp_file)
    for output in list_outputs:
        if output not in tablename:
            continue
        bucket = list_outputs[output]["bucket"]
        prefix = list_outputs[output]["prefix"]
        conn_id = list_outputs[output]["connection_id"]
        if not prefix.endswith("/"):
            prefix += "/"
        file_key = prefix + tmp_file[tmp_file.rfind("/")+1:]
        # Se sube el archivo al S3
        logger.info(f"Tamaño del archivo a subir: {size} bytes")
        if not save_df_to_s3(tmp_file, conn_id, bucket, file_key, provider, in_memory=False):
            raise AssertionError(f"No se guardó el archivo en la ruta indicada")
        # Se borra el archivo al finalizar el upload
        delete_temp_dir(tmp_file)
        break
    else:
        raise AssertionError(f"No se encontró la ruta de salida para {tablename}")
    end_process_datetime = datetime_by_tzone(timezone).strftime('%d/%m/%Y %H:%M:%S')
    task.xcom_push(key="END_PROCESS_DATETIME_" + str(task.map_index), value=end_process_datetime)


@task(task_id="MASTER_GENERATION", trigger_rule='none_skipped')
def get_generate_from_xcom(**kwargs):
    task = kwargs['ti']
    final_outputs = []
    conf = task.xcom_pull(task_ids="VALIDATE_TRANSFORMATION", key="CONTROL-CONFIG")
    tasks = get_tasks_from_control(conf, "generator")
    tasks_with_save = []
    definition = task.xcom_pull(task_ids="SCRIPTS-EXTRACTOR", key="EXTRACTION-DEFINITION-JSON")

    # LOGICA - TABLAS QUE SI TIENEN EL PARAMETRO SAVE_OUTPUT
    for item in definition:
        if "save_output" in item.keys() and item["save_output"]:
            tasks_with_save.append(item["identifier"])
    success_tasks = tasks["tasks"]
    success_tasks = [item[1] for item in success_tasks]
    logger.info(f"GENERADORES QUE FUERON EXITOSOS (TABLAS): {success_tasks}")
    xcom_keys = task.xcom_pull(task_ids="SCRIPTS-EXTRACTOR", key="XCOM-EXTRACTION-NAMES")
    logger.debug(xcom_keys)
    for key in xcom_keys:
        if not key.startswith(OperationTypeEnum.SELECT.value) and not key.startswith(OperationTypeEnum.PROCEDURE.value):
            continue
        xcom_outputs = task.xcom_pull(task_ids="SCRIPTS-EXTRACTOR", key=key)
        logger.info(f"Trayendo tablas {xcom_outputs}")
        for select in xcom_outputs:
            tablename = select_multiple(select)["tablename"]
            if (tasks["reset"] or tasks["status"] == ProcessStatusEnum.SUCCESS.value or select not in success_tasks) and tablename in tasks_with_save:
                final_outputs.append(select)
    Variable.set(key='GENERATES', value=final_outputs, serialize_json=True)
    if len(final_outputs) > 0:
        return [[item] for item in final_outputs]
    else:
        return [[None]]


def get_generate_task_group(db_intern_conn, parameters: Dict[str, Any], control_s3: Dict[str, Any],
                            timezone: str, provider: str) -> TaskGroup or None:
    group = None
    try:
        with TaskGroup(group_id="GeneracionyDespliegueDeResultados", prefix_group_id=False) as group:
            outputs = get_generate_from_xcom()

            tasks = PythonOperator.partial(
                task_id="GENERATORS",
                python_callable=generate_and_deploy,
                on_failure_callback=on_failure_generator,
                on_success_callback=on_success_generator,
                op_kwargs={'intern_conn': db_intern_conn, 'params': parameters, 'timezone': timezone,
                           'provider': provider}
            ).expand(op_args=outputs)

            validate_task = PythonOperator(
                task_id="VALIDATE_GENERATOR",
                python_callable=validate_generate,
                op_kwargs={'control_params': control_s3, 'timezone': timezone, 'provider': provider},
                trigger_rule='none_skipped'
            )
            outputs >> tasks >> validate_task
    except Exception as e:
        logger.error(f"Error creando taskGroup de generadores. {e}")
    finally:
        return group
