import json
from typing import Dict, Any
from airflow.utils.task_group import TaskGroup
from airflow.operators.python import PythonOperator
from airflow.models import Variable
from airflow.decorators import task
from airflow.exceptions import AirflowSkipException
from components.DatabaseOperation.DatabaseTransformation import delete_table, delete_procedure
from components.S3Route import get_files_from_prefix, get_file_from_prefix
from components.Xcom import save_commands_to_xcom
from components.Utils import update_sql_commands_2
from components.Control import get_tasks_from_control, update_new_process
from components.S3Route import load_control_to_s3
from components.Xcom import delete_all_xcom_tasks, delete_task_instances
from enums.ProcessStatusEnum import ProcessStatusEnum
from enums.OperationTypeEnum import OperationTypeEnum
from components.Timezone import datetime_by_tzone

import logging

logger = logging.getLogger()


def validate_transform(control_params: Dict[str, Any], timezone: str, provider: str, **kwargs) -> None:
    delete_task_instances()
    ti = kwargs["ti"]
    created_tables = ti.xcom_pull(task_ids="MASTER_TRANSFORMATION", key="TABLES_CREATED")
    lista = []
    if created_tables:
        for i in created_tables:
            lista.append(i)
    success_tasks = ti.xcom_pull(task_ids="TRANSFORMATIONS", key="SUCCESS_TASKS")
    failed_tasks = ti.xcom_pull(task_ids="TRANSFORMATIONS", key="FAILED_TASKS")
    conf = ti.xcom_pull(task_ids="VALIDATE_EXTRACTION", key="CONTROL-CONFIG")
    final_dict = {}
    status = ProcessStatusEnum.SUCCESS.value
    if not isinstance(success_tasks, type(None)) and len(success_tasks) > 0:
        for success_task in success_tasks:
            task = ti.xcom_pull(task_ids="TRANSFORMATIONS", key=success_task)[0]
            final_dict.update({success_task: task})
    if not isinstance(failed_tasks, type(None)) and len(failed_tasks) > 0:
        status = ProcessStatusEnum.FAIL.value
        for failed_task in failed_tasks:
            task = ti.xcom_pull(task_ids="TRANSFORMATIONS", key=failed_task)[0]
            final_dict.update({failed_task: task})
    conf = update_new_process(conf, status, final_dict, timezone, ti, lista, True)
    if status == ProcessStatusEnum.FAIL.value:
        conn = control_params["connection_id"]
        bucket = control_params["bucket"]
        prefix = control_params["prefix"]
        if not prefix.endswith("/"):
            prefix += "/"
        key = prefix + control_params["filename"]
        conf = json.dumps(conf, indent=2, default=str)
        loaded = load_control_to_s3(bytes(conf.encode()), conn, bucket, key, provider, timezone)
        if loaded:
            logger.info(f"Cargado correctamente el archivo de control en {key}")
        delete_all_xcom_tasks()
        raise AirflowSkipException(f"Ocurrieron errores en la etapa de transformación")
    elif status == ProcessStatusEnum.SUCCESS.value:
        ti.xcom_push(key="CONTROL-CONFIG", value=conf)


def on_failure_transform(context) -> None:
    ti = context["ti"]
    task_name = f"{ti.task_id}_{ti.map_index}"
    transform = Variable.get('TRANSFORMS', default_var=[], deserialize_json=True)
    script = transform[ti.map_index][0]
    exception = str(context["exception"])
    status = ProcessStatusEnum.FAIL.value
    init_process = ti.xcom_pull(task_ids="TRANSFORMATIONS", key="INIT_PROCESS_DATETIME_" + str(ti.map_index))[0]
    task_result = {"description": script, "status": status, "message": exception, "init_process": init_process,
                   "end_process": ""}
    ti.xcom_push(key=task_name, value=task_result)
    ti.xcom_push(key="FAILED_TASKS", value=task_name)


def on_success_transform(context) -> None:
    ti = context["ti"]
    task_name = f"{ti.task_id}_{ti.map_index}"
    transform = Variable.get('TRANSFORMS', default_var=[], deserialize_json=True)
    script = transform[ti.map_index][0]
    status = ProcessStatusEnum.SUCCESS.value
    init_process = ti.xcom_pull(task_ids="TRANSFORMATIONS", key="INIT_PROCESS_DATETIME_" + str(ti.map_index))[0]
    end_process = ti.xcom_pull(task_ids="TRANSFORMATIONS", key="END_PROCESS_DATETIME_" + str(ti.map_index))[0]
    task_result = {"description": script, "status": status, "message": "", "init_process": init_process,
                   "end_process": end_process}
    ti.xcom_push(key=task_name, value=task_result)
    ti.xcom_push(key="SUCCESS_TASKS", value=task_name)


def transformations(xcom_commands: str, intern_conn, timezone: str, **kwargs):
    if isinstance(xcom_commands, type(None)):
        raise AirflowSkipException
    task = kwargs['ti']
    init_process_datetime = datetime_by_tzone(timezone).strftime('%d/%m/%Y %H:%M:%S')
    task.xcom_push(key="INIT_PROCESS_DATETIME_" + str(task.map_index), value=init_process_datetime)
    engine = intern_conn.engine
    script_name = xcom_commands[0]
    commands = xcom_commands[1]
    logger.info(f"Ejecutando transformaciones del script {script_name}")
    not_procedure = ["UPDATE","SELECT","CREATE","ALTER","DROP","DELETE","INSERT","GRANT","REVOKE","TRUNCATE","COPY",
                     "COMMIT","ROLLBACK","USE","BEGIN"]
    with engine.connect() as connection:
        for command in commands:
            if any(command.startswith(palabra) or command.startswith(palabra.lower()) for palabra in not_procedure):
                logger.info(f"Ejecutando comando de transformación: {command}")
                _ = connection.execute(command)
            else:
                logger.info(f"Generando llamada al procedure según bd para: {command}")
                command = intern_conn.generate_sql_procedure(command)
                logger.info(f"EJECUTANDO FINAL PROCEDURE: {command}")
                _ = connection.execute(command)
    end_process_datetime = datetime_by_tzone(timezone).strftime('%d/%m/%Y %H:%M:%S')
    task.xcom_push(key="END_PROCESS_DATETIME_" + str(task.map_index), value=end_process_datetime)


@task(task_id="MASTER_TRANSFORMATION", trigger_rule='none_skipped')
def get_trans_from_xcom(provider, store_procedure, procedure_mask, label_tablename, transform_mask, order_delimiter,
                        db_intern_conn, **kwargs):
    task = kwargs['ti']
    #CAMBIOS PARA TRAER LOS SP
    proc_created = []
    tables_created = task.xcom_pull(task_ids="VALIDATE_EXTRACTION", key="TABLES_CREATED")
    lista = []
    for i in tables_created:
        lista.append(i)
    engine = db_intern_conn.engine
    conn_id = store_procedure["s3_params"]["connection_id"]
    bucket = store_procedure["s3_params"]["bucket"]
    prefix = store_procedure["s3_params"]["prefix"]
    procedures = get_files_from_prefix(conn_id, bucket, prefix, provider)
    definitions = task.xcom_pull(task_ids="SCRIPTS-EXTRACTOR", key="EXTRACTION-DEFINITION-JSON")
    #VALIDAR NOMBRE DE LOS PROCEDURES Y LABEL STORE
    for element in definitions:
        if "temp_table" in element.keys() and element["temp_table"] == True:
            lista.append(element["identifier"])

    with engine.connect() as connection:
        for procedure in procedures:
            procedure = procedure[1]
            if procedure.find(label_tablename + ":") != -1:
                index = procedure.find(label_tablename + ":")
                label_lenght = len(label_tablename + ":")
                tablename = procedure[index + label_lenght:].strip().split("\n")[0]
                delete_procedure(tablename, engine)
                for element in definitions:
                    if element["identifier"] == tablename and "transformation_store_procedure" in element.keys() and element["transformation_store_procedure"] == True:
                        logger.info(f"Ejecutando creacion de procedure: {procedure}")
                        # result = db_intern_conn.check_procedure(tablename, connection)
                        # if not result:
                        try:
                            _ = connection.execute(procedure)
                        except Exception as e:
                            logger.error(f" Error: {e}")
                            raise AirflowSkipException
                        proc_created.append(tablename)
                        lista.append(tablename)
            else:
                logger.debug(f"No se encontró el label en {procedure} por ende no se creará")
    save_commands_to_xcom(procedures, kwargs['ti'], procedure_mask, transform_mask, procedure_mask, order_delimiter)
    logger.debug(f"Procedures cargados en Xcom: {procedures}")

    transforms_per_file = []
    conf = task.xcom_pull(task_ids="VALIDATE_EXTRACTION", key="CONTROL-CONFIG")
    tasks = get_tasks_from_control(conf, "transformation")
    success_tasks = tasks["tasks"]
    success_tasks = [item[1] for item in success_tasks]
    logger.info(f"SCRIPTS QUE FUERON EXITOSOS: {success_tasks}")
    xcom_keys = task.xcom_pull(task_ids="SCRIPTS-EXTRACTOR", key="XCOM-EXTRACTION-NAMES")
    logger.debug(xcom_keys)
    for key in xcom_keys:
        final_transforms = []
        if not key.startswith(OperationTypeEnum.TRANSFORM.value):
            continue
        xcom_transforms = task.xcom_pull(task_ids="SCRIPTS-EXTRACTOR", key=key)
        logger.info(f"Trayendo comandos {xcom_transforms}")
        if tasks["reset"] or tasks["status"] == ProcessStatusEnum.SUCCESS.value or key not in success_tasks:
            for transform in xcom_transforms:
                final_transforms.append(transform)
            transforms_per_file.append((key, final_transforms))
    logger.info(f"Scripts para la transformación: {transforms_per_file}")
    task.xcom_push(key="PROC_CREATED", value=proc_created)
    task.xcom_push(key="TABLES_CREATED", value=lista)
    Variable.set(key='TRANSFORMS', value=transforms_per_file, serialize_json=True)
    if len(transforms_per_file) > 0:
        return [[item] for item in transforms_per_file]
    else:
        return [[None]]


def get_transform_task_group(db_intern_conn, timezone: str, control_s3: Dict[str, Any],
                             provider: str, store_procedure: Dict, procedure_mask: str,
                             label_tablename: str, transform_mask: str, order_delimiter: str) -> TaskGroup or None:
    group = None
    try:
        with TaskGroup(group_id="TransformacionDeDatos", prefix_group_id=False) as group:
            transforms = get_trans_from_xcom(provider, store_procedure, procedure_mask, label_tablename, transform_mask, order_delimiter,
                                             db_intern_conn)

            tasks = PythonOperator.partial(
                task_id="TRANSFORMATIONS",
                python_callable=transformations,
                op_kwargs={'intern_conn': db_intern_conn, 'timezone': timezone},
                on_failure_callback=on_failure_transform,
                on_success_callback=on_success_transform
            ).expand(op_args=transforms)

            validate_task = PythonOperator(
                task_id="VALIDATE_TRANSFORMATION",
                python_callable=validate_transform,
                op_kwargs={'control_params': control_s3, 'timezone': timezone, 'provider': provider},
                trigger_rule='none_skipped'
            )
            transforms >> tasks >> validate_task
    except Exception as e:
        logger.error(f"Error creando taskGroup de transformación. {e}")
    finally:
        return group
